# Copyright 2015 The Chromium OS Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

from __future__ import print_function

import base64
import fcntl
import json
import logging
import os
import signal
import struct
import termios
import threading

from ws4py.websocket import WebSocket

import factory_common  # pylint: disable=W0611
from cros.factory.test.utils.web_socket_utils import WebSocketHandshake


_SHELL = os.getenv('SHELL', '/bin/bash')
_BUFSIZ = 8192
_CONTROL_START = 128
_CONTROL_END = 129


class TerminalManager(object):
  """Object to manage Terminal service for goofy."""
  def handle_web_socket(self, request):
    if not WebSocketHandshake(request):
      return

    pid, fd = os.forkpty()
    if pid == 0:
      env = os.environ.copy()
      env['USER'] = os.getenv('USER', 'root')
      env['HOME'] = os.getenv('HOME', '/root')
      os.chdir(env['HOME'])
      os.execve(_SHELL, [_SHELL], env)

    class TerminalWebSocket(WebSocket):
      def __init__(self, fd, **kwargs):
        super(TerminalWebSocket, self).__init__(**kwargs)
        self._fd = fd
        self._control_state = None
        self._control_string = ''

      def HandlePTYControl(self, fd, control_string):
        msg = json.loads(control_string)
        command = msg['command']
        params = msg['params']
        if command == 'resize':
          # some error happened on websocket
          if len(params) != 2:
            return
          winsize = struct.pack('HHHH', params[0], params[1], 0, 0)
          fcntl.ioctl(fd, termios.TIOCSWINSZ, winsize)
        else:
          logging.warn('Invalid request command "%s"', command)

      def received_message(self, message):
        if message.is_text:
          if self._control_state:
            self._control_string += str(message)
          else:
            os.write(self._fd, str(message))
        elif message.is_binary:
          # The control section is
          # binary(_CONTROL_START)-text-binary(_CONTROL_END)
          # So we only can receive length 1 binary message one time.
          if len(message) != 1:
            # skip it
            logging.warn('Len of binary message is %d, not 1.', len(message))
            return
          if self._control_state:
            if chr(_CONTROL_END) == message.data[0]:
              self.HandlePTYControl(self._fd, self._control_string)
              self._control_state = None
              self._control_string = ''
            else:
              logging.warn('Unexpected control message %d', message.data[0])
          else:
            if chr(_CONTROL_START) == message.data[0]:
              self._control_state = _CONTROL_START
            else:
              logging.warn('Unexpected control message %d', message.data[0])

    ws = TerminalWebSocket(fd, sock=request.connection)
    t = threading.Thread(target=ws.run)
    t.daemon = True
    t.start()

    while True:
      try:
        ws.send(base64.b64encode(os.read(fd, _BUFSIZ)))
      except OSError:
        break

    os.kill(pid, signal.SIGTERM)
    os.wait()
